#!/usr/bin/env python3

from __future__ import annotations

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.6.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


import configparser
import json
import pathlib
import sys
from dataclasses import dataclass
from pprint import pprint
from typing import Dict, Union


@dataclass
class SitePackagesManager:
    site_packages_path: pathlib.Path

    @classmethod
    def from_sys_path(cls) -> SitePackagesManager:
        """
        Creates a new instance of the SitePackagesManager class from the system path.

        Returns:
            SitePackagesManager: A new instance of the SitePackagesManager class.
        """
        return cls(pathlib.Path(SitePackagesManager.get_site_packages_dir()))

    @staticmethod
    def get_site_packages_dir() -> str:
        """
        Returns the directory path for the site-packages.

        Returns:
            str: The path to the site-packages directory.
        """
        return next(p for p in sys.path if "site-packages" in p)

    def get_package_name(self, dist_info_dir: pathlib.Path) -> str:
        """
        Retrieves the package name from the METADATA file within a .dist-info directory.

        Args:
            dist_info_dir (pathlib.Path): The path to the .dist-info directory.

        Returns:
            str: The name of the package.
        """
        metadata_file = dist_info_dir / "METADATA"
        if metadata_file.exists():
            with metadata_file.open() as f:
                for line in f:
                    if line.startswith("Name:"):
                        return line.split(":", 1)[1].strip()
        return ""

    def find_console_scripts(self) -> Dict[str, Dict[str, str]]:
        """
        Returns a dictionary mapping package names to their console script names and corresponding commands.

        Returns:
            Dict[str, Dict[str, str]]: A dictionary where keys are package names and values are dictionaries
            mapping console script names to their corresponding commands.
        """
        scripts: Dict[str, Dict[str, str]] = {}

        for dist_info_dir in self.site_packages_path.glob("*.dist-info"):
            package_name = self.get_package_name(dist_info_dir)
            entry_points_file = dist_info_dir / "entry_points.txt"
            if entry_points_file.exists():
                config = configparser.ConfigParser()
                config.read(entry_points_file)
                if "console_scripts" in config.sections():
                    for name, value in config.items("console_scripts"):
                        scripts.setdefault(package_name, {})[name] = value

        return scripts

    def save_console_scripts_json(self, json_file: pathlib.Path) -> None:
        """
        Saves the console scripts dictionary to a JSON file in the site-packages directory.

        Args:
            scripts (Dict[str, str]): The dictionary of console scripts.
        """
        metadata = {"console_scripts": self.find_console_scripts()}
        with json_file.open("w") as outfile:
            json.dump(metadata, outfile, indent=4, default=str)


def main(json_file: Union[str, None]) -> None:
    manager = SitePackagesManager.from_sys_path()
    if json_file is None:
        pprint(manager.find_console_scripts())
    else:
        manager.save_console_scripts_json(pathlib.Path(json_file))


if __name__ == "__main__":
    if len(sys.argv) == 1:
        main(None)
    else:
        main(sys.argv[1])



#!/usr/bin/env python3

from __future__ import annotations

import difflib

from uvpipx.internal_libs.Logger import get_logger
from uvpipx.uvpipx_venv_load import uvpipx_load_venv

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.6.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


from typing import Union

from uvpipx.internal_libs.misc import (
    Elapser,
)

# TODO review code at this point


def upgrade(
    package_name: str,
    *,
    name_override: Union[None, str] = None,
) -> None:
    logger = get_logger("upgrade")

    venv_model, venv = uvpipx_load_venv(package_name, name_override)
    logger.log_info(f"⬆️  Upgrade {package_name}\n")

    old_vers = sorted(venv.installed_package())
    package_name_spec = [venv_model.main_package.package_name_spec]
    inject_name_spec = [
        inj.package_name_spec for inj in venv_model.injected_packages.values()
    ]
    upd_name_spec = " ".join(package_name_spec + inject_name_spec)

    with Elapser() as ela:
        venv.install(package_name_spec + inject_name_spec, allow_upgrade=True)
    logger.log_info(
        ela.ela_str(
            f" 📥 uv pip install {upd_name_spec} in uvpipx venv {venv_model.venv.name()}",
        ),
    )

    logger.log_info(
        f" 🟢 uvpipx venv {venv.venv_path.name} with {venv_model.main_package.package_name} ready",
    )

    new_vers = sorted(venv.installed_package())

    diff_vers = [
        n
        for n in difflib.ndiff([str(s) for s in old_vers], [str(s) for s in new_vers])
        if n[:2] in ["+ ", "- "]
    ]

    if diff_vers:
        logger.log_info("\n 🏗️  changes")
        logger.log_info("\n".join(f"   {n}" for n in diff_vers))
    else:
        logger.log_info("\n ⭕ no change")

    logger.log_info("")

    # with (pck_venv / "uvpipx.json").open("w") as outfile:
    #     json.dump(uvpipx_dict, outfile, indent=4, default=str)

    # logger.log_info(" 🎯 Re-Exposing program change")
    # relink_bins(package_name, expose_bin_names=expose_bin_names_, venv_name=venv_name)



from __future__ import annotations

import sys
from pathlib import Path
from typing import List, Union

from uvpipx.internal_libs.misc import cmd_prepare_env, cmd_run
from uvpipx.uvpipx_venv_load import uvpipx_load_venv


def run_venv_bin(
    package_name: str,
    cmdline: List[str],
    *,
    name_override: Union[None, str] = None,
) -> None:
    """venv is specific to uvpipx. it can replace inject, runpip (oups runuv), uninject

    in fact runpip could be replace by uvpipx venv "truc" uv -- pip install me
    """
    venv_model, venv = uvpipx_load_venv(package_name, name_override)

    env = cmd_prepare_env(None)
    env["VIRTUAL_ENV"] = str(venv.venv_path)
    env["PATH"] = str(venv.venv_path / ".venv/bin") + ":" + env["PATH"]
    if "PYTHONHOME" in env:
        del env["PYTHONHOME"]

    rc, stdo, stde = cmd_run(
        Path.cwd(),
        cmdline,
        env=env,
        raw_pipe=True,
    )

    # print(stdo)
    # print(stde, file=sys.stderr)

    sys.exit(rc)



from __future__ import annotations

from dataclasses import asdict
from pathlib import Path
from typing import Any, Dict, Literal, Tuple, Union

from uvpipx import uvpipx_venv_factory
from uvpipx.UvPipxModels import (
    UvPipxExposedModel,
    UvPipxExposeInstallSets,
    UvPipxModel,
    UvPipxPackageModel,
    UvPipxVenvExposeAppModel,
)


def check_and_upgrade(
    config: Dict[str, Any],
) -> (
    Tuple[Dict[str, Any], Literal[False]]
    | Tuple[dict[str, Any], Literal[True]]
    | Tuple[None, Literal[False]]
):
    vers = config.get("config_version")

    if vers == "0.2.0":
        return config, False

    if vers is None:
        transform = transform_old_to_0_2_0(config)
        if transform:
            return asdict(transform), True

    return None, False


def transform_old_to_0_2_0(old_config: dict) -> Union[UvPipxModel, None]:
    vers = old_config.get("version")

    if vers is None:
        venv_model, venv = uvpipx_venv_factory.uvpipx_venv_factory(
            old_config["package_name"],
            old_config["venv_name"],
        )

        install_sets = [
            UvPipxExposeInstallSets(
                [old_config["package_name"]],
                old_config["bin_names"],
            ),
        ]

        exposed_bins = {}
        for bin_pair in old_config["exposed_bins"]:
            venv_bin_split = bin_pair[0].rsplit("/", maxsplit=1)
            local_bin_split = bin_pair[1].rsplit("/", maxsplit=1)
            ren_bin = (
                local_bin_split[1] if local_bin_split[1] != venv_bin_split[1] else None
            )

            pl = uvpipx_venv_factory.path_link_factory(
                Path(bin_pair[0]),
                Path(local_bin_split[0]),
                rename_local_bin=ren_bin,
            )

            exposed_bins[venv_bin_split[1]] = UvPipxVenvExposeAppModel(
                venv_bin_split[1],
                str(pl.link_path),
                install_sets[0].package_name_sets,
            )

        injected_packages = {
            k: UvPipxPackageModel(
                package_name_spec=v,
                package_name=k,
            )
            for k, v in old_config.get("injected_package", {}).items()
        }

        install_sets.append(UvPipxExposeInstallSets(list(injected_packages.keys()), []))

        exposed = UvPipxExposedModel(
            str(venv.venv_path / ".venv/bin"),
            install_sets,
            exposed_bins,
        )

        return UvPipxModel(
            venv=venv_model,
            main_package=UvPipxPackageModel(
                package_name_spec=old_config["package_name_ref"],
                package_name=old_config["package_name"],
            ),
            injected_packages=injected_packages,
            exposed=exposed,
        )

    return None


# TODO move to test

# no_vers_config = {
#     "package_name_ref": "jc",
#     "package_name": "jc",
#     "venv_name": None,
#     "bin_names": ["*"],
#     "uvpipx_package_path": "/home/gaetan/.local/uv-pipx/venvs/jc",
#     "exposed_bins": [
#         [
#             "/home/gaetan/.local/uv-pipx/venvs/jc/.venv/bin/jc",
#             "/home/gaetan/.local/bin/jc",
#         ]
#     ],
#     "injected_package": {"rich": "rich==12.0.0"},
# }

# new_vers = transform_old_to_0_2_0(no_vers_config)
# pass



from __future__ import annotations

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.2.0"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


import os
import platform
import shutil
import subprocess  # nosec: B404  # noqa: S404
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Tuple, Type, TypeVar, Union, get_args, get_origin

from uvpipx.internal_libs.Logger import Logger, get_logger


# Définir le chemin du répertoire à parcourir
def find_executable(dir_path: Path, allow_symlink: bool = False) -> List[Path]:
    return [
        file
        for file in dir_path.iterdir()
        if file.is_file()
        and os.access(file, os.X_OK)
        and (not file.is_symlink() or allow_symlink)
    ]


@dataclass
class Elapser:
    """
    A context manager for measuring elapsed time.

    Explanation:
    This class can be used as a context manager to measure the elapsed time within the context block.

    Args:
        None

    Returns:
        None

    Examples:
        with Elapser() as elapser:
            # Code block to measure elapsed time
    """

    start: Union[None, float] = None
    end: Union[None, float] = None
    interval_seconds: float = -1
    elapsed_second: str = ""

    def __enter__(self) -> Elapser:
        self.start = time.perf_counter()
        return self

    def __exit__(self, *args: list) -> Union[bool, None]:  # , *args
        self.end = time.perf_counter()
        if self.start is None or self.end is None:
            msg = "Missing start or end"
            raise RuntimeError(msg)

        self.interval_seconds = self.end - self.start
        self.elapsed_second = f"{self.interval_seconds:.3f} seconds"
        return None

    def ela_str(self, message: str) -> str:
        return f"{message}   ⏱️  {self.elapsed_second}"


def shell_run(
    command: str,
    *,
    cwd: Union[None, Path] = None,
    env: Union[None, Dict[str, str]] = None,
    raise_on_error: bool = True,
) -> Tuple[int, Union[str], Union[str]]:
    """
    Executes a shell command and returns the result.

    Explanation:
    This function runs a shell command and captures the standard output, standard error, and return code.

    Args:
        command (List[str]): The shell command to be executed.
        raise_on_error (bool): Flag to raise an error if the command execution fails. Default is False.

    Returns:
        Tuple[int, str, str]: A tuple containing the return code, standard output, and standard error.
    """
    env_ = cmd_prepare_env(env)
    # encoding = cmd_prepare_encoding()

    opt_args = {}
    if cwd is not None:
        opt_args["cwd"] = cwd

    with subprocess.Popen(
        command,  # type: ignore[arg-type, call-overload]
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,  # noqa: S602
        text=True,
        env=env_,
        **opt_args,
    ) as proc:  # nosec: B602
        try:
            stdout, stderr = proc.communicate()
        except subprocess.SubprocessError:
            if raise_on_error:
                raise

        rc = proc.returncode
        # stdout = stdout.decode(encoding)
        # stderr = stderr.decode(encoding)
        if rc != 0 and raise_on_error:
            short_msg = f"{stderr:2000}".rstrip()
            msg = f"🔴 Command failed with return code {rc} {short_msg}..."
            raise RuntimeError(
                msg,
            )

    return rc, stdout, stderr


def cmd_run(
    cwd: Union[Path, str],
    command: Union[str, List[str]],
    *,
    env: Union[None, Dict[str, str]] = None,
    raise_on_error: bool = True,
    raw_pipe: bool = False,
) -> Tuple[Union[int, Any], Union[None, str], Union[None, str]]:
    """
    Executes a shell command and returns the result.

    Explanation:
    This function runs a shell command and captures the standard output, standard error, and return code.

    Args:
        command (List[str]): The shell command to be executed.
        raise_on_error (bool): Flag to raise an error if the command execution fails. Default is False.

    Returns:
        Tuple[int, str, str]: A tuple containing the return code, standard output, and standard error.
    """
    env_ = cmd_prepare_env(env)
    # encoding = cmd_prepare_encoding()
    pipe_type = None if raw_pipe else subprocess.PIPE

    with subprocess.Popen(
        command,
        stdout=pipe_type,
        stderr=pipe_type,
        cwd=cwd,
        shell=True,  # nosec: B602 # noqa: S602
        text=True,
        env=env_,
    ) as proc:
        try:
            if raw_pipe:
                stdout, stderr = (None, None)
                proc.wait()
            else:
                stdout, stderr = proc.communicate()
                # stdout = stdout.decode(encoding)
                # stderr = stderr.decode(encoding)
        except subprocess.SubprocessError:
            if raise_on_error:
                raise

    rc = proc.returncode
    if rc != 0 and raise_on_error:
        short_msg = f"{stderr:2000}".rstrip()
        msg = f"🔴 Command failed with return code {rc} {short_msg}..."
        raise RuntimeError(msg)

    return rc, stdout, stderr


def cmd_prepare_env(env: Union[None, Dict[str, str]]) -> Dict[str, str]:
    env_ = env
    if env_ is None:
        env_ = os.environ.copy()
        k_to_del = [k for k in env_ if "VIRTUAL_ENV" in k]
        for k in k_to_del:
            del env_[k]
    return env_


def cmd_prepare_encoding() -> str:
    os_type = platform.system()
    encoding = "utf-8"
    if os_type == "Windows":
        encoding = "utf-16"
    #     if "powershell" in command[0]:
    #         # command = ["powershell", "-Command"] + command
    #         encoding = 'utf-16'
    #     else:
    #         # command = ["cmd", "/c"] + command
    #         encoding = 'oem'
    return encoding


def shell_run_elapse(
    command: str,
    message: str,
    *,
    raise_on_error: bool = True,
    logger: Union[None, Logger] = None,
) -> None:
    logger_ = logger or get_logger("shell_run_elapse")
    with Elapser() as ela:
        rc, std_o, std_e = shell_run(command, raise_on_error=raise_on_error)

    logger_.log_info(f"{message}   ⏱️  {ela.elapsed_second}")


def command_exists(cmd: str) -> bool:
    return shutil.which(cmd) is not None


class InvalidTypeError(Exception):
    def __init__(self, expected_types: List[Type], actual_type: Type) -> None:
        self.expected_types = expected_types
        self.actual_type = actual_type
        super().__init__(self.__str__())

    def __str__(self) -> str:
        expected = ", ".join([t.__name__ for t in self.expected_types])
        return f"Invalid type: expected one of ({expected}), got {self.actual_type.__name__}"


# def check_type(value: Any, expected_types: List[Union[Type, None]]) -> None:
#     if not any(isinstance(value, t) for t in expected_types):
#         raise InvalidTypeError(expected_types, type(value))

T = TypeVar("T")


def check_type(value: Any, expected_types: Union[Type[T], List[Type[T]]]) -> T:  # noqa: ANN401
    expected_types_ = (
        expected_types if isinstance(expected_types, list) else [expected_types]
    )
    if value is None:
        raise InvalidTypeError(expected_types_, type(value))

    for expected_type in expected_types_:
        origin = get_origin(expected_type)
        if origin:
            args = get_args(expected_type)
            if isinstance(value, origin) and all(
                isinstance(elem, args[0]) for elem in value
            ):
                return value

        if isinstance(value, expected_type):
            return value

    raise InvalidTypeError(expected_types_, type(value))


def check_type_n_None(
    value: Any,  # noqa: ANN401
    expected_types: Union[Type[T], List[Type[T]]],
) -> Union[T, None]:
    expected_types_ = (
        expected_types if isinstance(expected_types, list) else [expected_types]
    )
    if value is None:
        return value

    try:
        for expected_type in expected_types_:
            origin = get_origin(expected_type)
            if origin:
                args = get_args(expected_type)
                if isinstance(value, origin) and all(
                    isinstance(elem, args[0]) for elem in value
                ):
                    return value

            if isinstance(value, expected_type):
                return value
    except Exception as e:
        raise InvalidTypeError(expected_types_, type(value)) from e

    raise InvalidTypeError(expected_types_, type(value))



from __future__ import annotations

import datetime
import os
from dataclasses import dataclass, field
from enum import Enum, IntEnum
from typing import Dict, List

from uvpipx.internal_libs.stylist import Color, Painter


class LogMode(Enum):
    PRINT = "print"
    BUFFER = "buffer"


class LogLevel(IntEnum):
    DEBUG = 0
    INFO = 50
    NOTICE = 100
    WARN = 200
    ERROR = 300


@dataclass
class LogEntry:
    level: LogLevel
    ts: datetime.datetime
    message: str


@dataclass
class Logger:
    log_mode: LogMode = LogMode.PRINT
    show_level: LogLevel = LogLevel.INFO
    show_log_mode_prefix: bool = os.getenv("UVPIPX_SHOW_LOG_PREFIX", "0").lower() in [
        "1",
        "true",
    ]
    __buffer: List[LogEntry] = field(init=False)

    def __post_init__(self) -> None:
        self.__buffer = []

    def render(self, messages: List[LogEntry]) -> None:
        for message in messages:
            if message.level >= self.show_level:
                if self.log_mode == LogMode.PRINT:
                    self.screen_log_entry(message)
                elif self.log_mode == LogMode.BUFFER:
                    self.__buffer.append(message)

    def render_level_prefix(self, level: LogLevel) -> str:
        text_info = ""
        if level == level.INFO:
            text_info = Painter.color_str("INFO", Color.BRIGHT_CYAN, Color.ST_BOLD)
        elif level == level.WARN:
            text_info = Painter.color_str("WARN", Color.BRIGHT_YELLOW, Color.ST_BOLD)
        elif level == level.ERROR:
            text_info = Painter.color_str("ERR ", Color.BRIGHT_RED, Color.ST_BOLD)
        elif level == level.DEBUG:
            text_info = Painter.color_str("DEBG", Color.BLUE)

        return f"[{text_info}] " if self.show_log_mode_prefix else ""

    def screen_log_entry(self, entry: LogEntry) -> None:
        print(f"{self.render_level_prefix(entry.level)}{entry.message}")

    def log_at_level(self, level: LogLevel, messages: str) -> None:
        ts = datetime.datetime.now().astimezone()
        log_messages = [
            LogEntry(level, ts, message) for message in messages.split("\n")
        ]
        self.render(log_messages)

    def log_debug(self, messages: str) -> None:
        """
        Logs a message with an optional verbosity level.

        Args:
            message (str): The message to be logged.
            level (int, optional): The verbosity level of the message. Defaults to 1.

        Returns:
            None
        """

        self.log_at_level(LogLevel.DEBUG, messages)

    def log_info(self, messages: str) -> None:
        """
        Logs an informational message.

        Explanation:
        This function prints an informational message with a specific format to the render.

        Args:
            message (str): The message to be logged.

        Returns:
            None
        """

        self.log_at_level(LogLevel.INFO, messages)

    def log_notice(self, messages: str) -> None:
        """
        Logs an notice message.

        Explanation:
        This function prints an notice message with a specific format to the render.

        Args:
            message (str): The message to be logged.

        Returns:
            None
        """

        self.log_at_level(LogLevel.INFO, messages)

    def log_warn(self, messages: str) -> None:
        """
        Logs a warm message.

        Args:
            message (str): The warm message to be logged.

        Returns:
            None
        """

        self.log_at_level(LogLevel.WARN, messages)

    def log_error(self, messages: str) -> None:
        """
        Logs a error message.

        Args:
            message (str): The error message to be logged.

        Returns:
            None
        """

        self.log_at_level(LogLevel.ERROR, messages)


LOGGER: Dict[str, Logger] = {}


def get_logger(name: str = "default") -> Logger:
    if name not in LOGGER:
        LOGGER[name] = Logger()
    return LOGGER[name]



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import List, Union

RE_PIP_REQ = re.compile(
    r"""(?P<name>[A-Za-z0-9._-]+)\s*                   # Nom du projet
(\[(?P<extras>[A-Za-z0-9._,-]+)\])?\s*        # Extras optionnels, entourés de []
(?P<version_specifiers>[^;]+)?\s*              # Spécifications des versions
(;\s*(?P<environment_marker>.+))?              # Environnement optionnel, après ;
""",
    re.VERBOSE,
)


@dataclass
class Requirement:
    # PEP 508 - Dependency specification for Python Software Packages
    name: str
    extras: Union[None, List[str]] = None
    version_specifiers: Union[None, List[str]] = None
    environment_marker: Union[None, str] = None

    @staticmethod
    def _split_str(data: str) -> List[str]:
        return [element.strip() for element in data.split(",")] if data else []

    @classmethod
    def from_str(cls, line: str) -> "Requirement":  # noqa: ANN102
        match = RE_PIP_REQ.match(line)
        if not match:
            msg = f"Line {line} not match PIP_REQ"
            raise RuntimeError(msg)

        return cls(
            name=match.group("name"),
            extras=Requirement._split_str(match.group("extras")),
            version_specifiers=Requirement._split_str(
                match.group("version_specifiers"),
            ),
            environment_marker=match.group("environment_marker"),
        )

    def to_str(self) -> str:
        s = f"{self.name}"
        if self.extras:
            s += f"[{','.join(self.extras)}]"
        if self.version_specifiers:
            s += f" {','.join(self.version_specifiers)}"
        if self.environment_marker:
            s += f"; {self.environment_marker}"

        return s


# requirements = [
#     "SomeProject",
#     "SomeProject == 1.3",
#     "SomeProject >= 1.2, < 2.0",
#     "SomeProject [foo, bar]",
#     "SomeProject ~= 1.4.2",
#     "SomeProject == 5.4 ; python_version < '3.8'",
#     "SomeProject ; sys_platform == 'win32'",
#     'requests [security] >= 2.8.1, == 2.8.* ; python_version < "2.7"',
#     "SomeProject==1.3",
#     "SomeProject>= 1.2, < 2.0",
#     "SomeProject[foo, bar]",
#     "SomeProject~=1.4.2",
#     "SomeProject==5.4;python_version<'3.8'",
#     "SomeProject;sys_platform=='win32'",
#     'requests[security]>=2.8.1,==2.8.*;python_version<"2.7"',
# ]



from __future__ import annotations

import sys

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.6.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


import os
from pathlib import Path
from typing import Union

# $XDG_CACHE_HOME = $HOME/.cache
# PIPX_HOME= so default to /home/gaetan/.local/pipx
# PIPX_LOCAL_VENVS=/home/gaetan/.local/pipx/venvs
# PIPX_VENV_CACHEDIR=/home/gaetan/.local/pipx/.cache
# PIPX_DEFAULT_PYTHON=/usr/bin/python3
# USE_EMOJI=true


def env_to_path(env_name: str, default: Union[str, Path, None] = None) -> Path:
    pat_s = os.environ.get(env_name, default)
    if pat_s is None:
        # if fail_is_none:
        msg = f"Unable to get value or default value for os env {env_name}"
        raise ValueError(msg)
        # return None

    pat_e = os.path.expandvars(pat_s)
    # check_type(p, [str])
    return Path(pat_e)


ci_project_dir = env_to_path("CI_PROJECT_DIR", ".")
home = Path(
    os.path.expanduser("~"),  # noqa: PTH111
)  # remark Path.expanduser("~") not work
uvpipx_home = env_to_path("UVPIPX_HOME", home / ".local/uv-pipx")
uvpipx_venvs = env_to_path("UVPIPX_LOCAL_VENVS", uvpipx_home / "venvs")

default_local_bin = home / ".local/bin"
if home == Path("/root"):
    default_local_bin = Path("/usr/local/bin")

uvpipx_local_bin = env_to_path("UVPIPX_BIN_DIR", default_local_bin)

platform = "win" if sys.platform.startswith("win") else sys.platform
uvpipx_self_dir = Path(__file__).resolve().parent.parent


# TODO on windows .venv/bin is .venv/Scripts so need to change expose



__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.6.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"

import os
import sys
from typing import List

import uvpipx.uvpipx_run
from uvpipx import (
    uvpipx_all,
    uvpipx_expose,
    uvpipx_infos,
    uvpipx_inject,
    uvpipx_install,
    uvpipx_upgrade,
)
from uvpipx.internal_libs.args import ArgParser
from uvpipx.internal_libs.Logger import get_logger
from uvpipx.internal_libs.misc import Elapser, check_type, check_type_n_None


def common_args(argp: ArgParser) -> None:
    logger = get_logger()
    argp.parse(sys.argv[2:])
    os.environ["UVPIPX_SHOW_DEBUG_LEVEL"] = str(
        argp.args["--verbose"].defaulted_value(),
    )
    logger.log_debug(f"Received args {sys.argv}")

    if argp.args["--help"].defaulted_value():
        argp.print_help()
        sys.exit(0)


def install(argp: ArgParser) -> None:
    """install package locally in their own venv"""
    logger = get_logger("install")

    common_args(argp)

    with Elapser() as ela:
        expose_rule_names = check_type_n_None(
            argp.args["--expose"].defaulted_value(),
            List[str],
        )
        force_reinstall = check_type(argp.args["--force"].defaulted_value(), bool)
        uvpipx_install.install(
            argp.args["python_pkg"].value,
            expose_rule_names=expose_rule_names,
            inject_pkgs=argp.args["--inject"].value,
            force_reinstall=force_reinstall,
        )

    logger.log_info(f"\n 🏁 Finish install  ⏱️  {ela.elapsed_second}")


def ensurepath(argp: ArgParser) -> None:
    """help to define PATH"""

    common_args(argp)

    uvpipx_infos.ensurepath()


def uvpipx_list(argp: ArgParser) -> None:
    """show the list of uvpipx venv"""

    common_args(argp)

    uvpipx_infos.uvpipx_list()


def info(
    argp: ArgParser,
) -> None:
    """show the list of uvpipx venv"""

    common_args(argp)

    get_venv = check_type(argp.args["--get-venv"].defaulted_value(), bool)
    uvpipx_infos.info(argp.args["python_pkg"].value, get_venv=get_venv)


def uninstall(argp: ArgParser) -> None:
    """uninstall package and their venv"""
    logger = get_logger("uninstall")

    common_args(argp)

    with Elapser() as ela:
        uvpipx_install.uninstall(
            argp.args["python_pkg"].value,
        )

    logger.log_info(f"\n 🏁 Finish uninstall  ⏱️  {ela.elapsed_second}")


def uninstall_all(argp: ArgParser) -> None:
    """upgrade all package locally in their own venv"""
    logger = get_logger("upgrade")

    common_args(argp)

    with Elapser() as ela:
        uvpipx_all.uninstall_all()

    logger.log_info(f"\n 🏁 Finish uninstall all ⏱️  {ela.elapsed_second}")


def venv(argp: ArgParser) -> None:
    """run command of a package without installing it"""
    common_args(argp)

    uvpipx.uvpipx_run.run_venv_bin(argp.args["python_pkg"].value, argp.extra_args)


def upgrade(argp: ArgParser) -> None:
    """upgrade package locally in their own venv"""
    logger = get_logger("upgrade")

    common_args(argp)

    with Elapser() as ela:
        uvpipx_upgrade.upgrade(
            argp.args["python_pkg"].value,
        )

    logger.log_info(f"\n 🏁 Finish upgrade  ⏱️  {ela.elapsed_second}")


def upgrade_all(argp: ArgParser) -> None:
    """upgrade all package locally in their own venv"""
    logger = get_logger("upgrade")

    common_args(argp)

    with Elapser() as ela:
        uvpipx_all.upgrade_all()

    logger.log_info(f"\n 🏁 Finish upgrade all  ⏱️  {ela.elapsed_second}")


def inject(argp: ArgParser) -> None:
    """inject package locally in venv"""
    logger = get_logger("inject")

    common_args(argp)

    with Elapser() as ela:
        inject_pkg = [argp.args["inject_python_pkg"].value, *argp.extra_args]
        uvpipx_inject.inject(argp.args["python_pkg"].value, inject_pkg)

    logger.log_info(f"\n 🏁 Finish inject  ⏱️  {ela.elapsed_second}")


def uninject(argp: ArgParser) -> None:
    """inject package locally in venv"""
    logger = get_logger("uninject")

    common_args(argp)

    with Elapser() as ela:
        inject_pkg = [argp.args["uninject_python_pkg"].value, *argp.extra_args]
        uvpipx_inject.uninject(argp.args["python_pkg"].value, inject_pkg)

    logger.log_info(f"\n 🏁 Finish uninject  ⏱️  {ela.elapsed_second}")


def expose(argp: ArgParser) -> None:
    """expose package locally in venv"""
    logger = get_logger("expose")

    common_args(argp)

    with Elapser() as ela:
        uvpipx_expose.expose(
            argp.args["python_pkg"].value,
            expose_rule_names=[argp.args["expose_rule_names"].value],
        )

    logger.log_info(f"\n 🏁 Finish expose  ⏱️  {ela.elapsed_second}")


def expose_all(argp: ArgParser) -> None:
    """expose all package locally in venv"""
    logger = get_logger("expose")

    common_args(argp)

    with Elapser() as ela:
        uvpipx_expose.expose_all([argp.args["expose_rule_names"].value])

    logger.log_info(f"\n 🏁 Finish expose all  ⏱️  {ela.elapsed_second}")



from __future__ import annotations

from uvpipx.internal_libs.Logger import get_logger
from uvpipx.uvpipx_core import UvPipxVenv
from uvpipx.uvpipx_venv_factory import path_link_from_model
from uvpipx.uvpipx_venv_load import uvpipx_load_venv
from uvpipx.UvPipxModels import UvPipxModel
from uvpipx.version import show_version

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.6.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


import os
from pathlib import Path
from typing import List, Union

from uvpipx import config
from uvpipx.internal_libs.misc import shell_run


def ensurepath() -> None:
    logger = get_logger("uninstall")

    path_set = os.environ["PATH"].split(":")
    if str(config.uvpipx_local_bin) + "x" in path_set:
        logger.log_info("🟢 Configuration of PATH already OK")
        return

    profile_path = Path(os.environ["HOME"]) / ".profile"

    profile_content = Path(profile_path).read_text()
    if "# Added by uvpipx" in profile_content:
        logger.log_info(f"""⚠️  Configuration already in {profile_path}

⚠️  To use without restart the shell, launch 
export PATH=$PATH:{config.uvpipx_local_bin}
""")
        return

    line_to_add = ""
    if Path(profile_path).read_text()[-1] != "\n":
        line_to_add = "\n"

    line_to_add += f"""# Added by uvpipx
export PATH=$PATH:{config.uvpipx_local_bin}\n"""

    with profile_path.open("a") as file:
        file.write(line_to_add)

    logger.log_info(f"""Configuration added to {profile_path}

⚠️  To use without restart the shell, launch 
export PATH=$PATH:{config.uvpipx_local_bin}
""")


def _info(uvpipx: UvPipxModel, venv: UvPipxVenv) -> str:
    def get_version(pck_name: str, stdout_line: List[str]) -> str:
        vers = next(
            (line for line in stdout_line if pck_name in line),
            None,
        )
        if vers is None:
            vers = f"{pck_name} Unknow version"

        return vers

    rc, stdout, stderr = shell_run("uv pip freeze", cwd=venv.venv_path)
    stdout_line = stdout.split("\n")
    main_vers = get_version(uvpipx.main_package.package_name, stdout_line)

    injected_vers = []
    for pkg_injected in list(uvpipx.injected_packages.keys()):
        injected_vers.append(get_version(pkg_injected, stdout_line))
    injected_vers = sorted(injected_vers)

    path_links = (
        [path_link_from_model(uvpipx.exposed, m) for m in uvpipx.exposed.apps.values()]
        if uvpipx.exposed and uvpipx.exposed.apps
        else []
    )
    bins = "\n".join(f"   ✅ {app_bin.show_name_with_link()}" for app_bin in path_links)

    # TODO add show info about install set in advanced view

    if not bins:
        bins = "   ❌ Nothing exposed"

    output = f""" 📦 {uvpipx.main_package.package_name} ({main_vers}) in venv {uvpipx.venv.uvpipx_dir}"""
    if injected_vers:
        output_inject = "\n".join(f"""   📦 {pkg_ver}""" for pkg_ver in injected_vers)
        output += f"""

 🎯 Injected packages
{output_inject}"""

    output += f"""

 🎯 Exposed program
{bins}\n"""

    return output


def _info_pkg(
    package_name: str,
    *,
    name_override: Union[None, str] = None,
    get_venv: bool = False,
) -> str:
    uvpipx, venv = uvpipx_load_venv(package_name, name_override)

    return str(venv.venv_path / ".venv") if get_venv else _info(uvpipx, venv)


def info(
    package_name: str,
    *,
    name_override: Union[None, str] = None,
    get_venv: bool = False,
) -> None:
    logger = get_logger("info")

    info = _info_pkg(package_name, name_override=name_override, get_venv=get_venv)
    logger.log_info(info)


def uvpipx_list() -> None:
    infos = f"""📍 uvpipx venvs are in {config.uvpipx_venvs}
💻 apps are exposed at {config.uvpipx_local_bin} 

"""
    logger = get_logger("uvpipx_list")

    show_version()

    nb = 0
    if config.uvpipx_venvs.exists():
        for pck_venv in config.uvpipx_venvs.iterdir():
            infos += _info_pkg(pck_venv.name)  # this is a tricky way to get the name
            infos += "\n\n"
            nb += 1

    if nb == 0:
        infos += "⭕ No uvpipx package installed !"

    logger.log_info(infos)



from __future__ import annotations

import json
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Any, Dict, List, TypeVar, Union

T = TypeVar("T")


@dataclass
class UvPipxVenvExposeAppModel:
    bin_app_name: str
    exposed_app_path: str
    packages_name_sets: List[str]

    def exposed_app(self) -> Path:
        return Path(self.exposed_app_path)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "UvPipxVenvExposeAppModel":
        return UvPipxVenvExposeAppModel(
            bin_app_name=data["bin_app_name"],
            exposed_app_path=data["exposed_app_path"],
            packages_name_sets=data["packages_name_sets"],
        )


@dataclass
class UvPipxExposeInstallSets:
    package_name_sets: List[str]
    exposed_apps_rules: List[str]

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "UvPipxExposeInstallSets":
        return UvPipxExposeInstallSets(
            package_name_sets=data["package_name_sets"],
            exposed_apps_rules=data["exposed_apps_rules"],
        )


@dataclass
class UvPipxExposedModel:
    venv_bin_dir: str
    install_sets: List[UvPipxExposeInstallSets] = field(default_factory=list)
    apps: Dict[str, UvPipxVenvExposeAppModel] = field(default_factory=dict)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "UvPipxExposedModel":
        return UvPipxExposedModel(
            venv_bin_dir=data["venv_bin_dir"],
            install_sets=[
                UvPipxExposeInstallSets.from_dict(item) for item in data["install_sets"]
            ],
            apps={
                k: UvPipxVenvExposeAppModel.from_dict(v)
                for k, v in data["apps"].items()
            },
        )


@dataclass
class UvPipxVenvModel:
    uvpipx_dir: str
    name_override: Union[None, str] = None

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "UvPipxVenvModel":
        return UvPipxVenvModel(
            uvpipx_dir=data["uvpipx_dir"],
            name_override=data.get("name_override"),
        )

    def name(self) -> str:
        return self.name_override or self.uvpipx_path().name

    def uvpipx_path(self) -> Path:
        return Path(self.uvpipx_dir)


@dataclass
class UvPipxPackageModel:
    package_name_spec: str
    package_name: str

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "UvPipxPackageModel":
        return UvPipxPackageModel(
            package_name_spec=data["package_name_spec"],
            package_name=data["package_name"],
        )


@dataclass
class UvPipxModel:
    venv: UvPipxVenvModel
    main_package: UvPipxPackageModel
    injected_packages: Dict[str, UvPipxPackageModel]
    exposed: Union[UvPipxExposedModel, None]
    config_version: str = "0.2.0"

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "UvPipxModel":
        return UvPipxModel(
            venv=UvPipxVenvModel.from_dict(data["venv"]),
            main_package=UvPipxPackageModel.from_dict(data["main_package"]),
            injected_packages={
                k: UvPipxPackageModel.from_dict(v)
                for k, v in data["injected_packages"].items()
            },
            exposed=UvPipxExposedModel.from_dict(data["exposed"]),
            config_version=data.get("config_version", "0.2.0"),
        )

    def save_json(self, file_name: Union[str, Path]) -> None:
        with (self.venv.uvpipx_path() / file_name).open("w") as outfile:
            json.dump(to_dict(self), outfile, indent=4, default=str)


def to_dict(obj: Any) -> Dict[str, Any]:  # noqa: ANN401
    """
    A helper function to recursively convert a dataclass instance to a dictionary.
    """
    return asdict(obj)


# cat /home/gaetan/.local/uv-pipx/venvs/jc/uvpipx.json


class UvPipVenvNotReady(Exception):
    def __init__(self, message: str, error_code: Union[None, int] = None) -> None:
        super().__init__(message)
        self.message = message
        self.error_code = error_code



from __future__ import annotations

import json
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Tuple, Union

from uvpipx import config
from uvpipx.internal_libs.Logger import Logger, get_logger
from uvpipx.uvpipx_core import UvPipxVenv
from uvpipx.uvpipx_venv_factory import path_link_factory
from uvpipx.uvpipx_venv_load import uvpipx_load_venv
from uvpipx.UvPipxModels import UvPipxExposedModel, UvPipxVenvExposeAppModel


@dataclass
class ExposeApps:
    venv: UvPipxVenv
    logger: Union[None, Logger] = None
    prev_exposed: Union[None, UvPipxExposedModel] = None

    def __post_init__(self) -> None:
        self.logger_ = self.logger or get_logger()

    def set_prev_exposed(self, prev_exposed: Union[None, UvPipxExposedModel]) -> None:
        self.prev_exposed = prev_exposed

    def get_apps_list(
        self,
        expose_app_rules: List[str],
        main_package_name: str,
    ) -> Tuple[List[Path], Dict[str, str]]:
        # expose_apps_list = [
        #     Path(e) if isinstance(e, str) else e for e in expose_app_rules
        # ]
        renamed_apps = {}

        if not expose_app_rules:
            return ([], {})

        expose_apps_list = None
        expose_fallback = None
        if expose_app_rules == ["__eponym__"]:
            if self.venv.venv_bin(main_package_name).exists():
                expose_apps_list = [self.venv.venv_bin(main_package_name)]
            else:
                expose_fallback = "__main__"
                self.logger_.log_warn(
                    f" ⚠️  no find eponym app {self.venv.venv_bin(main_package_name)}. will fallback to app of main package",
                )

        if expose_app_rules == ["__main__"] or expose_fallback == "__main__":
            self.venv.update_metadata()

            with (self.venv.venv_path / "pip_metadata.json").open("r") as infile:
                metadata_dict = json.load(infile)

            console_scripts = metadata_dict["console_scripts"].get(
                main_package_name, [],
            )
            if console_scripts:
                expose_apps_list = [
                    self.venv.venv_bin(main_bin)
                    for main_bin in console_scripts
                    if not re.match(r"^(python|pip)(\d+(\.\d+)?)?$", main_bin)
                ]
                expose_fallback = None
            else:
                expose_fallback = "__eponym__"
                self.logger_.log_warn(
                    " ⚠️  cannot find metadata of package. fallback to eponym app",
                )

        if expose_fallback == "__eponym__":
            if self.venv.venv_bin(main_package_name).exists():
                self.logger_.log_info(" 🔰 fallback to eponym app is OK ")
                expose_fallback = None
                expose_apps_list = [self.venv.venv_bin(main_package_name)]
            else:
                expose_fallback = "__all__"
                self.logger_.log_warn(
                    f" ⚠️  fallback also fail no find eponym app {self.venv.venv_bin(main_package_name)}. will fallback to all app in venv",
                )

        if expose_app_rules == ["__all__"] or expose_fallback == "__all__":
            expose_apps_list = self.venv.venv_bins(
                regex_to_exclude=[r"^(python|pip)(\d+(\.\d+)?)?$"],
            )
        elif not expose_apps_list:
            renamed_apps = {
                key: value
                for item in expose_app_rules
                if ":" in item
                for key, value in [item.split(":", 1)]
            }
            expose_apps_list = [
                self.venv.venv_bin(item.split(":", 1)[0]) for item in expose_app_rules
            ]

        return expose_apps_list, renamed_apps

    def get_apps_to_remove(
        self,
        future_exposed_apps: Dict[str, UvPipxVenvExposeAppModel],
    ) -> Dict[str, Path]:
        if self.prev_exposed is None:
            return {}

        prev_exposed_apps = {
            k: Path(app.exposed_app_path) for k, app in self.prev_exposed.apps.items()
        }
        # remove pre_exposed_apps that are not in future_exposed_apps
        return {
            k: v for k, v in prev_exposed_apps.items() if k not in future_exposed_apps
        }

    def add_exposing(
        self,
        exposing_apps: List[Path],
        renamed_apps: Dict[str, str],
        pkgs_sets: List[str],
    ) -> Dict[str, UvPipxVenvExposeAppModel]:
        config.uvpipx_local_bin.mkdir(parents=True, exist_ok=True)
        exposed_apps = {}

        for app_bin in exposing_apps:
            pl = path_link_factory(
                app_bin,
                config.uvpipx_local_bin,
                rename_local_bin=renamed_apps.get(app_bin.name),
            )
            if not pl.link_exists():
                self.logger_.log_info(
                    f" 🎯 Exposing program {pl.show_name_with_link()}",
                )
                pl.link()
                exposed_apps[app_bin.name] = UvPipxVenvExposeAppModel(
                    app_bin.name,
                    str(pl.link_path),
                    pkgs_sets,
                )
            elif pl.is_valid():
                self.logger_.log_info(
                    f" 🔵 Already exposed to current uvpipx venv program {pl.show_name_with_link()}",
                )
                exposed_apps[app_bin.name] = UvPipxVenvExposeAppModel(
                    app_bin.name,
                    str(pl.link_path),
                    pkgs_sets,
                )
            else:
                if pl.link_path:
                    self.logger_.log_warn(
                        f" ❌ Not exposing program {app_bin.name}; {pl.link_path} already exist to {pl.link_path.resolve()}!",
                    )
                else:
                    msg = "link_path is None"
                    raise ValueError(msg)

        return exposed_apps

    def remove_exposing(
        self,
        remove_apps: Dict[str, Path],
    ) -> None:
        for app_bin, link_path in remove_apps.items():
            if link_path.exists():
                self.logger_.log_info(
                    f" 🗑️  Removing exposing {link_path} -> {app_bin} ",
                )
                link_path.unlink()

    def expose(
        self,
        main_package_name: str,
        expose_app_rules: List[str],
        pkgs_sets: List[str],
    ) -> Dict[str, UvPipxVenvExposeAppModel]:
        exposing_apps, renamed_apps = self.get_apps_list(
            expose_app_rules,
            main_package_name,
        )

        exposed_apps = self.add_exposing(exposing_apps, renamed_apps, pkgs_sets)
        remove_apps = self.get_apps_to_remove(exposed_apps)
        self.remove_exposing(remove_apps)

        return exposed_apps

    # TODO add function to clean old link in local bin


def expose(package_name: str, expose_rule_names: List[str]) -> None:
    logger = get_logger("expose")

    logger.log_info(f"🔭  Exposing apps of {package_name}\n")

    expose_rule_names_ = expose_rule_names

    uvpipx_model, venv = uvpipx_load_venv(
        package_name,
    )
    expo_app = ExposeApps(venv, logger)
    expo_app.set_prev_exposed(uvpipx_model.exposed if uvpipx_model else None)
    package_name = uvpipx_model.main_package.package_name

    uvpipx_model.exposed.apps = expo_app.expose(
        package_name, expose_rule_names_, [package_name],
    )
    uvpipx_model.save_json("uvpipx.json")

    logger.log_info(
        f" 🟢 uvpipx venv {uvpipx_model.venv.name()} with {package_name} ready",
    )


def expose_all(expose_rule_names: List[str]) -> None:
    logger = get_logger("expose_all")

    infos = ""

    nb = 0
    if config.uvpipx_venvs.exists():
        for pck_venv in config.uvpipx_venvs.iterdir():
            if nb > 0:
                logger.log_info("")
            expose(
                pck_venv.name,
                expose_rule_names,
            )  # This is a tricky way to get the name
            logger.log_info(" ----------------")
            nb += 1

    if nb == 0:
        infos += "⭕ No uvpipx package installed !"

    if infos:
        logger.log_info(infos)



from __future__ import annotations

import json
from typing import Tuple, Union

from uvpipx.uvpipx_core import UvPipxVenv
from uvpipx.uvpipx_venv_factory import uvpipx_venv_factory
from uvpipx.UvPipxModels import UvPipVenvNotReady, UvPipxModel
from uvpipx.UvPipxModelsUpgrader import check_and_upgrade


def uvpipx_load_venv(
    package_name: str,
    name_override: Union[str, None] = None,
) -> Tuple[UvPipxModel, UvPipxVenv]:
    venv_m, venv = uvpipx_venv_factory(package_name, name_override)

    if not venv.exists():
        msg = f"🔴 {package_name} not exist (path {venv.venv_path})"
        raise UvPipVenvNotReady(msg)

    if not (venv.venv_path / "uvpipx.json").exists():
        msg = f"🔴 {package_name} not exist (path {venv.venv_path})"
        raise UvPipVenvNotReady(msg)

    with (venv.venv_path / "uvpipx.json").open("r") as infile:
        file_dict = json.load(
            infile,
        )

    current_vers_dict, upgraded = check_and_upgrade(file_dict)
    if current_vers_dict is None:
        msg = "Bad venv, please unistall and install"
        raise UvPipVenvNotReady(msg)

    uvpipx_cfg = UvPipxModel.from_dict(current_vers_dict)

    if upgraded:
        uvpipx_cfg.save_json("uvpipx.json")

    return uvpipx_cfg, venv



from __future__ import annotations

import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Union

from uvpipx import config
from uvpipx.internal_libs.misc import find_executable, shell_run
from uvpipx.req_spec import Requirement


@dataclass
class UvPipxVenv:
    venv_path: Path

    def exists(self) -> bool:
        return (self.venv_path / ".venv").exists()

    def create_venv_if_need(self) -> bool:
        self.venv_path.mkdir(exist_ok=True, parents=True)

        if not self.exists():
            rc, std_o, std_e = shell_run(
                f"uv venv {self.venv_path / '.venv'}",
                raise_on_error=True,
            )
            return True
        return False

    def freeze(self) -> str:
        rc, stdout, stderr = shell_run("uv pip freeze", cwd=self.venv_path)
        return stdout  # if isinstance(stdout, str) else stdout.decode("utf-8")

    def installed_package(self) -> list[str]:
        return self.freeze().rstrip().split("\n")

    def installed_package_as_req(self) -> list[Requirement]:
        req = self.freeze().rstrip().split("\n")
        req_tuple = [Requirement.from_str(r) for r in req]

        return req_tuple

    def install(self, packages_name_spec: List[str], allow_upgrade: bool = False) -> str:
        opt = " --upgrade" if allow_upgrade else ""
        install_pkgs = " ".join(packages_name_spec)
        rc, stdout, stderr = shell_run(
            f"uv pip install{opt} {install_pkgs}",
            cwd=self.venv_path,
        )
        return stdout  # if isinstance(stdout, str) else stdout.decode("utf-8")

    def uninstall(self, package_name_spec: str) -> str:
        rc, stdout, stderr = shell_run(
            f"uv pip uninstall {package_name_spec}",
            cwd=self.venv_path,
        )
        return stdout  # if isinstance(stdout, str) else stdout.decode("utf-8")

    def venv_bin_dir(self) -> Path:
        return self.venv_path / ".venv/bin"

    def venv_bins(self, regex_to_exclude: list[str]) -> list[Path]:
        re_comp_to_exclude = [re.compile(r) for r in regex_to_exclude]
        all_bins = find_executable(self.venv_bin_dir())
        return [
            s for s in all_bins if not any(r.search(s.name) for r in re_comp_to_exclude)
        ]

    def venv_bin(self, name: str, fail_if_notexist: bool = True) -> Path:
        path_ = self.venv_bin_dir() / name

        if fail_if_notexist and not path_.exists():
            msg = f"🔴 {name} not exist (path {path_})"
            raise RuntimeError(msg)

        return path_

    def run_in_venv(
        self,
        cmdline: str,
        *,
        cwd: Union[None, Path] = None,
        env: Union[None, Dict[str, str]] = None,
    ) -> None:
        rc, stdo, stde = shell_run(cmdline, cwd=cwd, env=env)
        if rc != 0:
            msg = f"🔴 Command failed with return code {rc} {stdo} {stde}"
            raise RuntimeError(msg)

    def update_metadata(self, if_not_exist: bool = True) -> None:
        pip_metadata = self.venv_path / "pip_metadata.json"
        if pip_metadata.exists() and if_not_exist:
            return

        uvpipx_console_scripts = (
            config.uvpipx_self_dir / "uvpipx/uvpipx_console_scripts.py"
        )
        python_venv_bin = self.venv_bin_dir() / "python"
        self.run_in_venv(
            f"{python_venv_bin} {uvpipx_console_scripts} {pip_metadata}",
        )

@dataclass
class PathLink:
    local_path: Path
    link_path: Union[Path, None] = None

    def exists(self) -> bool:
        return self.local_path.exists()

    def link_exists(self) -> bool:
        if self.link_path is None:
            msg = "Link not defined"
            raise RuntimeError(msg)

        return self.link_path.exists()

    def is_valid(self) -> bool:
        if self.link_path is None:
            msg = "link_path is None"
            raise ValueError(msg)

        if self.link_path.exists():
            return self.link_path.is_symlink() and (
                self.local_path == self.link_path.resolve()
            )

        return False

    def link(self) -> None:
        if self.link_path is None:
            msg = "link_path is None"
            raise ValueError(msg)

        os.symlink(self.local_path, self.link_path)

    def unlink(self, if_valid: bool = True) -> None:
        if self.link_path is None:
            msg = "link_path is None"
            raise ValueError(msg)

        if (if_valid and self.is_valid()) or (not if_valid and self.exists()):
            self.link_path.unlink()

    def show_name_with_link(self) -> str:
        if self.link_path is None:
            msg = "link_path is None"
            raise ValueError(msg)

        return (
            self.link_path.name
            if self.link_path.name == self.local_path.name
            else f" {self.local_path.name} -> {self.link_path.name}"
        )



#!/usr/bin/env python3

from __future__ import annotations

from uvpipx import config

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = "GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"
__version__ = "0.5.0"
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"

import shutil
from dataclasses import dataclass
from typing import List, Tuple, Union

from uvpipx.internal_libs.Logger import get_logger
from uvpipx.internal_libs.misc import Elapser
from uvpipx.req_spec import Requirement
from uvpipx.uvpipx_expose import ExposeApps
from uvpipx.uvpipx_venv_factory import path_link_from_model, uvpipx_venv_factory
from uvpipx.uvpipx_venv_load import uvpipx_load_venv
from uvpipx.UvPipxModels import (
    UvPipVenvNotReady,
    UvPipxExposedModel,
    UvPipxExposeInstallSets,
    UvPipxModel,
    UvPipxPackageModel,
)


@dataclass
class Installer:
    package_name_spec: str
    expose_rule_names: Union[None, List[str]] = None
    inject_pkgs_name_spec: Union[None, List[str]] = None
    name_override: Union[None, str] = None
    force_reinstall: bool = False

    def __post_init__(self):
        self.logger = get_logger("install")
        self.package_spec = Requirement.from_str(self.package_name_spec)
        self._inject_pkgs_name_spec = self.inject_pkgs_name_spec or []
        self.inject_pkgs_spec = [
            Requirement.from_str(pkg) for pkg in self._inject_pkgs_name_spec
        ]
        self.all_pkgs_name_spec = [self.package_name_spec, *self._inject_pkgs_name_spec]
        self.all_pkgs_spec = [self.package_spec, *self._inject_pkgs_name_spec]
        self.package_name = self.package_spec.name
        self.prepare_expose_rule_names()
        self.uvpipx_cfg = None
        self.venv_model = None
        self.venv = None

    def prepare_expose_rule_names(self)  -> None:
        self.expose_rule_names_def = self.expose_rule_names or ["__main__"]

    def check_existing_installation(self) -> Tuple:
        try:
            uvpipx_prev, venv_prev = uvpipx_load_venv(
                self.package_name,
                self.name_override,
            )
            if self.force_reinstall:
                if uvpipx_prev.exposed and len(uvpipx_prev.exposed.install_sets) > 1:
                    self.logger.log_warn(
                        f"⚠️  {self.package_name} already installed but in multiple steps (install/inject). Manual uninstall/install/inject or just try an upgrade is needed",
                    )
                    return uvpipx_prev, venv_prev, False
            else:
                self.logger.log_warn(
                    f"⚠️  {self.package_name} already installed. Use --force to reinstall from scratch",
                )
                return uvpipx_prev, venv_prev, False
        except UvPipVenvNotReady:
            return None, None, True

        return uvpipx_prev, venv_prev, True

    def create_virtual_env_if_needed(self) -> None:
        with Elapser() as ela:
            created = self.venv.create_venv_if_need()
        if created:
            self.logger.log_info(
                ela.ela_str(f" 📦 uv venv {self.venv_model.name()} created"),
            )

    def install_all_packages(self) -> None:
        with Elapser() as ela:
            self.venv.install(self.all_pkgs_name_spec)
        self.logger.log_info(
            ela.ela_str(
                f" 📥 uv pip install {self.all_pkgs_name_spec} in uvpipx venv {self.venv_model.name()}",
            ),
        )

    def save_pip_infos(self) -> None:
        (self.venv.venv_path / "requirements.txt").write_text(self.venv.freeze())
        pip_metadata = self.venv.venv_path / "pip_metadata.json"
        uvpipx_console_scripts = (
            config.uvpipx_self_dir / "uvpipx/uvpipx_console_scripts.py"
        )
        python_venv_bin = self.venv.venv_bin_dir() / "python"
        self.venv.run_in_venv(
            f"{python_venv_bin} {uvpipx_console_scripts} {pip_metadata}",
        )

    def expose_binaries(
        self,
        prev_exposed: Union[None, UvPipxExposedModel] = None,
    ) -> None:
        expo_app = ExposeApps(self.venv, self.logger)
        expo_app.set_prev_exposed(prev_exposed)
        main_install_set = UvPipxExposeInstallSets(
            [self.package_name],
            self.expose_rule_names_def,
        )
        exposed_bins = expo_app.expose(
            self.package_name,
            self.expose_rule_names_def,
            main_install_set.package_name_sets,
        )
        install_sets = [main_install_set]
        self.uvpipx_cfg.exposed = UvPipxExposedModel(
            str(self.venv.venv_path / ".venv/bin"),
            install_sets,
            exposed_bins,
        )

    def install(self) -> None:
        uvpipx_prev, _, can_install = self.check_existing_installation()

        if not can_install:
            return

        uvpipx_injected_package = {
            k.name: UvPipxPackageModel(
                k.to_str(),
                k.name,
            )
            for k in self.inject_pkgs_spec
        }

        self.venv_model, self.venv = uvpipx_venv_factory(
            self.package_name,
            self.name_override,
        )
        self.uvpipx_cfg = UvPipxModel(
            venv=self.venv_model,
            main_package=UvPipxPackageModel(self.package_name_spec, self.package_name),
            injected_packages=uvpipx_injected_package,
            exposed=UvPipxExposedModel(str(self.venv.venv_bin)),
        )

        self.create_virtual_env_if_needed()
        self.install_all_packages()

        self.save_pip_infos()
        self.logger.log_info("")
        self.expose_binaries(prev_exposed=uvpipx_prev.exposed if uvpipx_prev else None)
        self.uvpipx_cfg.save_json("uvpipx.json")

        self.logger.log_info(
            f" 🟢 uvpipx venv {self.venv_model.name()} with {self.package_name} ready",
        )


def install(
    package_name_spec: str,
    *,
    expose_rule_names: Union[None, List[str]] = None,
    inject_pkgs: Union[None, List[str]] = None,
    name_override: Union[None, str] = None,
    force_reinstall: bool = False,
) -> None:
    config = Installer(
        package_name_spec,
        expose_rule_names=expose_rule_names,
        inject_pkgs_name_spec=inject_pkgs,
        name_override=name_override,
        force_reinstall=force_reinstall,
    )
    config.install()


def uninstall(package_name: str, *, name_override: Union[None, str] = None) -> None:
    logger = get_logger("uninstall")
    uvpipx, venv = uvpipx_load_venv(package_name, name_override)
    logger.log_info(f"🪓 Uninstalling {package_name}\n")
    logger.log_info("🗑️  Remove exposed program")
    if uvpipx.exposed:
        for mpl in uvpipx.exposed.apps.values():
            pl = path_link_from_model(uvpipx.exposed, mpl)
            pl.unlink()
            logger.log_info(f" ❌ Remove Exposed program {pl.show_name_with_link()}")
    # TODO also delete injected exposed bin
    logger.log_info(f"\n🗑️  Remove uvpipx venv {venv.venv_path.name}")
    shutil.rmtree(venv.venv_path)



from __future__ import annotations

from pathlib import Path
from typing import Union

from uvpipx import config
from uvpipx.uvpipx_core import PathLink, UvPipxVenv
from uvpipx.UvPipxModels import (
    UvPipxExposedModel,
    UvPipxVenvExposeAppModel,
    UvPipxVenvModel,
)


def uvpipx_venv_factory(
    package_name: str,
    name_override: Union[str, None] = None,
) -> tuple[UvPipxVenvModel, UvPipxVenv]:
    name = name_override or package_name

    venv_m = UvPipxVenvModel(str(Path(config.uvpipx_venvs) / name), name_override)
    venv = UvPipxVenv(venv_path=venv_m.uvpipx_path())

    return venv_m, venv


def path_link_factory(
    app_bin: Path,
    local_app_bin_dir: Path,
    rename_local_bin: Union[None, str] = None,
) -> PathLink:
    expose_local_bin_name = local_app_bin_dir / (rename_local_bin or app_bin.name)
    p_link = PathLink(app_bin, expose_local_bin_name)

    return p_link


def path_link_from_model(
    exposed: UvPipxExposedModel,
    app_bin_model: UvPipxVenvExposeAppModel,
) -> PathLink:
    p_link = PathLink(
        Path(exposed.venv_bin_dir) / app_bin_model.bin_app_name,
        Path(app_bin_model.exposed_app_path),
    )

    return p_link



#!/usr/bin/env python3

from __future__ import annotations

from uvpipx.internal_libs.Logger import get_logger
from uvpipx.req_spec import Requirement
from uvpipx.uvpipx_venv_load import uvpipx_load_venv
from uvpipx.UvPipxModels import UvPipxExposeInstallSets, UvPipxPackageModel

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.6.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


from typing import Union

from uvpipx.internal_libs.misc import (
    Elapser,
)


def inject(
    package_main_name: str,
    lst_package_name_spec: list[str],
    *,
    # expose_bin_names: Optional[List[str]] = None,
    name_override: Union[None, str] = None,
) -> None:
    logger = get_logger("inject")

    uvpipx_cfg, venv = uvpipx_load_venv(package_main_name, name_override)

    injecting_package = {
        r.name: r for rl in lst_package_name_spec for r in [Requirement.from_str(rl)]
    }

    for pck_name in injecting_package:  # TODO allow upgrade
        if pck_name in uvpipx_cfg.injected_packages:
            msg = f"🔴 {pck_name} already injected"
            raise RuntimeError(msg)

    pip_packages_spec = " ".join(lst_package_name_spec)

    with Elapser() as ela:
        venv.install(lst_package_name_spec, allow_upgrade=False)
    logger.log_info(
        ela.ela_str(
            f" 📥 uv pip install {pip_packages_spec} in uvpipx venv {uvpipx_cfg.venv.name()}",
        ),
    )

    logger.log_info(f" 🟢 injected {lst_package_name_spec}")
    (venv.venv_path / "requirements.txt").write_text(venv.freeze())
    logger.log_info("")

    injected_package = {
        k: UvPipxPackageModel(
            injecting_package[k].to_str(),
            k,
        )
        for k in injecting_package
    }
    tmp_dict = {**uvpipx_cfg.injected_packages, **injected_package}
    uvpipx_cfg.injected_packages = {k: tmp_dict[k] for k in sorted(tmp_dict)}
    if uvpipx_cfg.exposed:
        uvpipx_cfg.exposed.install_sets.append(
            UvPipxExposeInstallSets(list(injected_package.keys()), []),
        )
    else:
        msg = "Cannot inject on empty venv"
        raise RuntimeError(msg)

    uvpipx_cfg.save_json("uvpipx.json")

    # TODO add exposing


def uninject(
    package_main_name: str,
    lst_package_name_spec: list[str],
    *,
    name_override: Union[None, str] = None,
) -> None:
    logger = get_logger("uninject")

    uvpipx_cfg, venv = uvpipx_load_venv(package_main_name, name_override)

    uninjected_package = {
        r.name: r for rl in lst_package_name_spec for r in [Requirement.from_str(rl)]
    }

    for pck_name in uninjected_package:
        if pck_name not in uvpipx_cfg.injected_packages:
            msg = f"🔴 {pck_name} is not injected"
            raise RuntimeError(msg)

    pip_packages_spec = " ".join(uninjected_package.keys())
    with Elapser() as ela:
        venv.uninstall(pip_packages_spec)
    logger.log_info(
        ela.ela_str(
            f" 🗑️  uv pip uninstall {pip_packages_spec} in uvpipx venv {uvpipx_cfg.venv.name()}",
        ),
    )

    logger.log_info(f" 🗑️  uninjected {lst_package_name_spec}")
    (venv.venv_path / "requirements.txt").write_text(venv.freeze())
    logger.log_info("")

    stay_injected_package = {
        pck_name: pck_ref
        for pck_name, pck_ref in uvpipx_cfg.injected_packages.items()
        if pck_name not in uninjected_package
    }

    uvpipx_cfg.injected_packages = {
        k: stay_injected_package[k] for k in sorted(stay_injected_package)
    }

    uvpipx_cfg.save_json("uvpipx.json")

    # TODO change exposing



from __future__ import annotations

from uvpipx.internal_libs.Logger import get_logger
from uvpipx.uvpipx_install import uninstall
from uvpipx.uvpipx_upgrade import upgrade

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.6.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


from uvpipx import config


def upgrade_all() -> None:
    logger = get_logger("upgrade_all")

    infos = ""

    nb = 0
    if config.uvpipx_venvs.exists():
        for pck_venv in config.uvpipx_venvs.iterdir():
            if nb > 0:
                logger.log_info("")
            upgrade(pck_venv.name)  # This is a tricky way to get the name
            logger.log_info(" ----------------")
            nb += 1

    if nb == 0:
        infos += "⭕ No uvpipx package installed !"

    if infos:
        logger.log_info(infos)


def uninstall_all() -> None:
    logger = get_logger("uninstall_all")

    infos = ""

    nb = 0
    if config.uvpipx_venvs.exists():
        for pck_venv in config.uvpipx_venvs.iterdir():
            if nb > 0:
                logger.log_info("")
            uninstall(pck_venv.name)  # This is a tricky way to get the name
            logger.log_info(" ----------------")
            nb += 1

    if nb == 0:
        infos += "⭕ No uvpipx package installed !"

    if infos:
        logger.log_info(infos)



from __future__ import annotations

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.2.0"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


import os
import re
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict


class Color(Enum):
    # Foreground colors
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"

    BRIGHT_RED = "\033[91m"
    BRIGHT_GREEN = "\033[92m"
    BRIGHT_YELLOW = "\033[93m"
    BRIGHT_BLUE = "\033[94m"
    BRIGHT_MAGENTA = "\033[95m"
    BRIGHT_CYAN = "\033[96m"
    BRIGHT_WHITE = "\033[97m"

    # Background colors
    BG_RED = "\033[41m"
    BG_GREEN = "\033[42m"
    BG_YELLOW = "\033[43m"
    BG_BLUE = "\033[44m"
    BG_MAGENTA = "\033[45m"
    BG_CYAN = "\033[46m"
    BG_WHITE = "\033[47m"

    BG_BRIGHT_RED = "\033[101m"
    BG_BRIGHT_GREEN = "\033[102m"
    BG_BRIGHT_YELLOW = "\033[103m"
    BG_BRIGHT_BLUE = "\033[104m"
    BG_BRIGHT_MAGENTA = "\033[105m"
    BG_BRIGHT_CYAN = "\033[106m"
    BG_BRIGHT_WHITE = "\033[107m"

    # Reset and text styles
    ST_RESET = "\033[0m"
    ST_BOLD = "\033[1m"
    ST_DIM = "\033[2m"
    ST_ITALIC = "\033[3m"
    ST_UNDERLINE = "\033[4m"
    ST_BLINK = "\033[5m"
    ST_REVERSE = "\033[7m"
    ST_HIDDEN = "\033[8m"

    @staticmethod
    def rgb_color(r: int, g: int, b: int, background: bool = False) -> str:
        """Generates an ANSI escape sequence for a given RGB color."""
        if background:
            return f"\033[48;2;{r};{g};{b}m"  # ANSI escape for background color
        else:
            return f"\033[38;2;{r};{g};{b}m"  # ANSI escape for foreground color


class Painter:
    @staticmethod
    def hex_color(hexcode: str, background: bool = False) -> str:
        """Generates an ANSI escape sequence from a hex color code."""
        if os.getenv("NO_ANSI_COLOR", "false").lower() not in ["false", "0"]:
            return ""

        hexcode = hexcode.strip("#")
        r, g, b = int(hexcode[:2], 16), int(hexcode[2:4], 16), int(hexcode[4:6], 16)
        return Color.rgb_color(r, g, b, background)

    @staticmethod
    def color_str(message: str, *styles: Color) -> str:
        if os.getenv("NO_ANSI_COLOR", "false").lower() not in ["false", "0"]:
            return message

        style_codes = "".join(style.value for style in styles)
        return f"{style_codes}{message}{Color.ST_RESET.value}"

    @staticmethod
    def parse_color_tags(message: str) -> str:
        pattern = re.compile(r"</?([A-Z_]+)>")
        styles_set = set()
        message_ = message

        # Function to process each match
        def replace_tag(tag: str) -> str:
            if os.getenv("NO_ANSI_COLOR", "false").lower() not in ["false", "0"]:
                return ""

            tag_ = tag[1:-1]
            if not tag_.startswith("/"):
                styles_set.add(tag_)
                return Color[tag_].value

            styles_set.remove(tag_[1:])
            ansi_style = Color["ST_RESET"].value
            for style in styles_set:
                ansi_style += Color[style].value

            return ansi_style

        match_tag = pattern.search(message_)
        while match_tag:
            m = match_tag[0]
            message_ = message_.replace(m, replace_tag(m), 1)
            match_tag = pattern.search(message_)

        return message_


class RenderEmoji(Enum):
    EMOJI = "emoji"
    STR = "str"
    REMOVE = "remove"


@dataclass
class Emoji:
    render_emoji: RenderEmoji = RenderEmoji.EMOJI
    emoji_to_str: Dict[str, str] = field(init=False)

    def __post_init(self) -> None:
        self.emoji_to_str = {
            "🔴": "",
            "🟢": "",
            "🎯": "",
            "⚠️": "",
            "📥": "",
            "🗑️": "",
            "🏗️": "",
            "⭕": "",
            "✅": "",
            "❌": "",
            "📦": "",
        }

    def r(self, emoji: str) -> str:
        if self.render_emoji == RenderEmoji.EMOJI:
            return emoji
        elif self.render_emoji == RenderEmoji.STR:
            return self.emoji_to_str.get(emoji, "")
        # elif self.render_emoji == RenderEmoji.REMOVE:
        #     return ""

        return ""
